/*
 * main.c
 *
 *  Created on: 15.12.2008
 *      Author: Ti
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <string.h>

#include <pthread.h>
#include <unistd.h>

#define THREAD_USE 10000		// When do use threads in normal program use
#define THREAD_MAX 64			// Max threads to create

int benchmark = 0, max_threads = THREAD_MAX, force_threads = 0, num_loops = 1;
clock_t time_diff, time_tmp;
float duration, duration_part;
char chars[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefighijklmnopqrstuvwxyz0123456789$!";

struct thread
{
	pthread_t thread;
	char *password;
	unsigned int seed;
	int start;
	int end;
};

int get_cpu_count();
void *make_chars(void *struct_thread);
void show_help();

int main(int argc, char **argv)
{
	// default password length
	unsigned int passwdl = 10;
	unsigned int seed = 0;
	char *password;
	int option, loop_count = 0, cpu_count = get_cpu_count();
	struct timeval t;
	struct thread *threads;

	while((option = getopt(argc, argv, "hbl:t:c:")) != -1)
	{
		switch(option)
		{
		case 'h':
			show_help();
			exit(EXIT_SUCCESS);
		case 'b':
			benchmark = 1;
			break;
		case 'l':
			passwdl = (unsigned int) atoi(optarg);
			break;
		case 'c':
			num_loops = atoi(optarg);
			break;
		case 't':
			max_threads = atoi(optarg);
			force_threads = 1;
			break;
		case '?':
			fprintf(stderr, "Unrecognized option\n");
			show_help();
			exit(EXIT_SUCCESS);
			break;
		}
	}

	if(max_threads > passwdl && force_threads)
	{
		printf("It makes no sense to start more threads than have chars to calculate!\n");
		exit(EXIT_SUCCESS);
	}

	if(max_threads > THREAD_MAX)
	{
		printf("Sorry, limit of max threads reached, or do you have >64 cores in your system?\n");
		exit(EXIT_SUCCESS);
	}

	password = (char *) malloc(sizeof(char) * passwdl + 1);
	if(password == NULL)
	{
		printf("Sorry, could not allocate such much memory! (%u bytes)\n", passwdl + 1);
		exit(EXIT_FAILURE);
	}

	// Intialize random seed from /dev/random and time() (only linux)
#ifdef __linux__
	FILE *pFile = fopen("/dev/urandom", "r");
	unsigned int i = 0;
	char buf [10];

	if(pFile!=NULL)
	{
		fgets(buf, sizeof(buf), pFile);
		fclose(pFile);
	}

	for(i=0; i < sizeof(buf); i++)
		seed += (unsigned int) buf[i];
#endif

	if(benchmark == 1)
	{
		printf("Benchmark mode: CPUs found: %i\n", cpu_count);
	}

	time_diff = clock();
	STARTLOOP : loop_count++;

	memset(password, 0, sizeof(char) * passwdl + 1);

	gettimeofday(&t, 0);
	seed += (unsigned int) t.tv_usec;

	// Use only one thread
	if(passwdl < THREAD_USE && force_threads == 0)
	{
		threads = (struct thread *) malloc(sizeof(struct thread));
		threads->seed = seed;
		threads->password = password;
		threads->start = 0;
		threads->end = passwdl;

		pthread_create(&threads->thread, NULL, make_chars, (void*) threads);
		pthread_join(threads->thread, NULL);

		free(threads);
	}
	// Use many threads
	else
	{
		int thread_num = 0, thread_size = 0, j = 0, k = 0, chc = passwdl, start = 0, end = 0;

		// Automatic thread settings
		thread_size = passwdl / cpu_count;
		thread_num = passwdl / thread_size;
		if(passwdl % thread_size != 0)
		{
			while(thread_num >= cpu_count)
			{
				thread_size++;
				thread_num = passwdl / thread_size;
			}

			thread_num++;
		}

		if(thread_num > max_threads)
		{
			thread_size = passwdl / max_threads;
			thread_num = passwdl / thread_size;
			thread_num++;
		}

		// Overwrite automatic thread settings
		if(force_threads)
		{
			thread_size = passwdl / max_threads;
			thread_num = passwdl / thread_size;

			if(passwdl % thread_size != 0)
			{
				while(thread_num >= max_threads)
				{
					thread_size++;
					thread_num = passwdl / thread_size;
				}

				thread_num++;
			}

			if(benchmark == 1)
				printf("Threads starting: --%02i--\n", thread_num);
		}

		threads = (struct thread *) malloc(sizeof(struct thread) * thread_num);

		for(j=0; chc > 0; j++)
		{
			if(chc > thread_size)
			{
				end = chc;
				start = chc - thread_size;
			}
			else
			{
				end = chc;
				start = 0;
			}

			chc -= thread_size;
			chc--;

			threads[j].seed = seed + chc;
			threads[j].password = password;
			threads[j].start = start;
			threads[j].end = end;

			if(benchmark == 1)
				printf("Thread %i: %i to %i\n", j, start, end);

			pthread_create(&threads[j].thread, NULL, make_chars, (void*) &threads[j]);
		}

		for(k=0; k < j; k++)
		{
			pthread_join(threads[k].thread, NULL);
		}

		free(threads);
	}

	time_tmp = clock() - time_diff;
	duration_part = duration;
	duration = (float) time_tmp / CLOCKS_PER_SEC;

	if(benchmark == 1)
	{
		printf("Generated %u chars in %f seconds (loop %i) total: %f\n", passwdl, duration - duration_part, loop_count, duration);
	}
	else
	{
		password[passwdl] = '\0';

		printf("Password: %s\n", password);
	}

	if(loop_count < num_loops)
		goto STARTLOOP;

	free(password);

	return EXIT_SUCCESS;
}

void *make_chars(void *struct_thread)
{
	unsigned int i = 0;
	int charlen = strlen(chars);
	char tmp;
	struct thread *my_thread = (struct thread *) struct_thread;

	srand(my_thread->seed);

	for(i=my_thread->start; i<=my_thread->end; i++)
	{
		tmp = chars[rand() % charlen];
		my_thread->password[i] = tmp;
	}

	return 0;
}

void show_help()
{
	printf("Use: passgen [-hb] [-l length] [-c loops] [-t threads]\n");
	printf("\t-h       Shows this help\n");
	printf("\t-b       Start in benchmark mode\n");
	printf("\t-l       Sets length of password to generate\n");
	printf("\t-c       Sets number of passwords to generate\n");
	printf("\t-t       Sets number of threads to start\n");
}

int get_cpu_count()
{
	int count = 1;

#ifdef __linux__
	char buf[10];
	FILE *pp = popen("cat /proc/cpuinfo | grep processor | wc -l", "r");
	fgets(buf, 5, pp);
	count = atoi(buf);
	fclose(pp);
#elif __WIN32__
	count = atoi(getenv("NUMBER_OF_PROCESSORS"));
#endif

	return count;
}
